
from collections import namedtuple
import formula_utilities as U

class STLError(Exception):
  pass



class LeafOp():
  def __init__(self):
    self.op = []
    self.param = 0

  def compute(self, x):
    if self.op:
      raise STLError("No operator is defined for LeafOp")

    if self.op == '>':
      return x - self.param
    if self.op == '<':
      return self.param - x





class Interval():

  def __init__(self, lb, ub):
    self.lb = lb
    self.ub = ub

  def __add__(self, other):
    n = Interval(0, 0)
    n.ub = self.ub + other.ub
    n.lb = self.lb + other.lb
    return n

  def __sub__(self, other):
    n = Interval(0, 0)
    n.ub = self.ub - other.lb
    n.lb = self.lb - other.ub
    return n

  def update_bounds(self, ub=None, lb=0):
    if self.lb < lb:
      self.lb = lb
    if ub and self.ub > ub:
      self.ub = ub


DataPoint = namedtuple('DataPoint', ['time', 'value'])

class SyntaxTreeNode():

  def __init__(self):
    self.left_node = []
    self.right_node = []
    self.parent = []

    self.op = []
    self.metric = [] # index of the metric
    self.interval = []
    self.signal_length = 0

    self.data_points = [] # it will be a sorted list(?) of DataPoints according to the time, points from this list will
                          # be removed when dp.time < time - interval.ub
    self.data_points_to_add = [] # sorted list, the points from this list will be added when
                                 # dp.time <= time - interval.lb

    self.data_points_r = [] # only for S, data points from the right tree
    self.data_points_to_add_r = [] # only for S, data points to add for the right tree.
    self.qv = 0
    self.nt = 0


    # for the leaf nodes
    self.param = 0 # for now 1 dimensional signal

  def initialize_node(self, tokens, index):
    op = tokens[index]
    # it is an operand, so leaf node.
    if U.is_operand(op):
      self.metric = int(tokens[index][1:len(tokens[index])])
      self.op = tokens[index + 1]
      self.param = int(tokens[index + 2])
      return index + 3
    else:
      index_shift = 1 # the operator
      self.op = op
      self_length = 0
      if op in U.STLOperators:
        self.interval = Interval(int(tokens[index+1]), int(tokens[index+2]))
        index_shift += 2 # interval vars
        self_length = self.interval.ub
      self.left_node = SyntaxTreeNode()
      post_index = self.left_node.initialize_node(tokens, index + index_shift)
      if U.is_binary(op):
        self.right_node = SyntaxTreeNode()
        post_index = self.right_node.initialize_node(tokens, post_index)
        self.signal_length = self_length + max(self.right_node.signal_length, self.left_node.signal_length)
      else:
        self.signal_length = self_length + self.left_node.signal_length
      return post_index

  def _get_until_interval(self, index, time_points, signal_length):

    if not time_points:
      new_interval_1 = Interval(0, self.interval.lb)
      new_interval_1 = self._get_interval(index, time_points, new_interval_1, False, signal_length)
      return new_interval_1, self._get_interval(index, time_points, self.interval, False, signal_length)

  def _get_since_interval(self, index, time_points, signal_length, t):

    if not time_points:
      new_interval_1 = Interval(0, self.interval.lb)
      new_interval_1 = self._get_interval(index, time_points, new_interval_1, True, signal_length)
      new_interval_2 = self._get_interval(index, time_points, self.interval, True, signal_length)
      return new_interval_1, new_interval_2

    i1 = Interval(0, self.interval.lb)
    r1 = self._get_interval(index, time_points, i1, True, signal_length, t)
    r2 = self._get_interval(index, time_points, self.interval, True, signal_length, t)
    return r1, r2


  def _get_interval(self, index, time_points, interval, is_past, signal_length=None, t=None):
    """Compute indices for the resulting interval. This function handles different cases.
     Case 1: time_points is None. In this case the interval is computed as interval +- index.

     Case 2: time_points array is given.
    """
    new_interval = Interval(index, index)
    if not time_points:
      if is_past:
        new_interval -= interval
      else:
        new_interval += interval
      new_interval.update_bounds(ub=signal_length)
      return new_interval

    # Return the indices according to the time points and current time t
    if not is_past:
      return None # TODO implement for the future operators

    time_interval = Interval(t, t)
    time_interval -= interval
    index_interval = Interval(0, index)
    # now find the corresponding indices, start from i:

    # Loop for upper bound:
    for x in xrange(index, -1, -1):
      if time_points[x] <= time_interval.ub:
        index_interval.ub = x
        break
    # Loop for lower bound:
    for x in xrange(index_interval.ub - 1, -1, -1):
      if time_points[x] <= time_interval.lb:
        index_interval.lb = x
        break

    return index_interval



  def compute(self, signal, index, time_points=None, t=None):
    if self.op == '&':
      return min(self.left_node.compute(signal, index, time_points, t),
                 self.right_node.compute(signal, index, time_points, t))
    if self.op == '|':
      return max(self.left_node.compute(signal, index, time_points, t),
                 self.right_node.compute(signal, index, time_points, t))
    if self.op == '!':
      return -1*self.left_node.compute(signal, index, time_points, t)

    # Future operators:
    if self.op == 'F':
      val = U.MIN_VAL

      r = self._get_interval(index, time_points, self.interval, False, len(signal))
      for x in xrange(r.lb, r.ub + 1):
        vc = self.left_node.compute(signal, x, time_points)
        val = vc if vc > val else val
      return val

    if self.op == 'G':
      val = U.MAX_VAL
      r = self._get_interval(index, time_points, self.interval, False, len(signal))
      for x in xrange(r.lb, r.ub + 1):
        ts = time_points[x] if time_points else None
        vc = self.left_node.compute(signal, x, time_points, ts)
        val = vc if vc < val else val
      return val

    if self.op == 'U':
      val = U.MIN_VAL
      min_phi1_val = U.MAX_VAL
      r1, r2 = self._get_until_interval(index, time_points, len(signal))
      for x in xrange(r1.lb, r1.ub):
        ts = time_points[x] if time_points else None
        min_phi1_val = min(min_phi1_val, self.left_node.compute(signal, x, time_points, ts))

      for x in xrange(r2.lb, r2.ub + 1):
        ts = time_points[x] if time_points else None
        phi2_val = self.right_node.compute(signal, x, time_points, ts)
        val = max(val, min(min_phi1_val, phi2_val))
        phi1_val = self.left_node.compute(signal, x, time_points, ts)
        if phi1_val < val:
          return val # we can stop the iteration, after this point the new value will always be less than val
        min_phi1_val = min(min_phi1_val, phi1_val)
      return val

    if self.op == 'P':
      val = U.MIN_VAL
      r = self._get_interval(index, time_points, self.interval, True, len(signal), t)
      for x in xrange(r.lb, r.ub + 1):
        ts = time_points[x] if time_points else None
        vc = self.left_node.compute(signal, x, time_points, ts)
        val = vc if vc > val else val
      return val

    if self.op == 'A':
      val = U.MAX_VAL
      r = self._get_interval(index, time_points, self.interval, True, len(signal), t)
      for x in xrange(r.lb, r.ub + 1): #xrange(max(index - self.interval.ub, 0), max(0, index - self.interval.lb + 1)):
        ts = time_points[x] if time_points else None
        vc = self.left_node.compute(signal, x, time_points, ts)
        val = vc if vc < val else val
      return val

    if self.op == 'S':
      val = U.MIN_VAL
      min_left = U.MAX_VAL
      r1, r2 = self._get_since_interval(index, time_points, len(signal), t)
      for x in xrange(r1.ub, r1.lb, -1):
        ts = time_points[x] if time_points else None
        t_left = self.left_node.compute(signal, x, time_points, ts)
        min_left = t_left if t_left < min_left else min_left
      pre_val = U.MIN_VAL
      for x in xrange(r2.ub, r2.lb - 1, -1):
        ts = time_points[x] if time_points else None
        t_right = self.right_node.compute(signal, x, time_points, ts)
        t_left = self.left_node.compute(signal, x, time_points, ts)
        min_left = min(t_left, min_left)
        val = min(t_right, min_left) if min(t_right, min_left) > val else val
        if val < pre_val:
          return pre_val
        else:
          pre_val = val

        # t_right = self.right_node.compute(signal, x, time_points)
        # val = min(t_right, min_left) if min(t_right, min_left) > val else val
        # t_left = self.left_node.compute(signal, x, time_points)
        # if t_left < val:
        #   return val
        # min_left = min(t_left,min_left)
      return val
    # Base case, lead computation
    if self.op == '>':
      return signal[index][self.metric] - self.param
    if self.op == '<':
      return self.param - signal[index][self.metric]


  def _remove_condition(self):
    if len(self.data_points) > 0:
      # self.data_points_to_add[0].value is the new point to add
      if self.op == 'P':
        return self.data_points[-1].value < self.data_points_to_add[0].value
        # keep a descending list for Previous op
      elif self.op == 'A':
        return self.data_points[-1].value > self.data_points_to_add[0].value
        # keep an ascending list for Always in the past op
    else:
      return False

  def _process_ordered_list(self, list_to_process, new_point, decreasing):
    last_dropped_point = None
    while len(list_to_process) > 0:
      if decreasing:
        if list_to_process[-1].value < new_point.value:
          last_dropped_point = list_to_process.pop()
        else:
          break
      else:
        if list_to_process[-1].value > new_point.value:
          last_dropped_point = list_to_process.pop()
        else:
          break
    if len(list_to_process) == 0 or list_to_process[-1].value != new_point.value:
      if last_dropped_point and len(list_to_process) > 0:
        list_to_process.append(DataPoint(time=last_dropped_point.time, value=new_point.value))
      else:
        list_to_process.append(new_point)
      # update the expiration

    return last_dropped_point

  def compute_nt(self, data_points, data_points_to_add, dropped_point):
    """Compute the expiration time, e.g., when the qv might change even if a new point has not arrived."""

    nt = U.MAX_VAL # if it is the only one, it will never expire
    if len(data_points) > 1:
      #if dropped_point:  # This should be used for the computation, change the second ones nt to this:
      #  data_points[-1] = DataPoint(time=dropped_point.time, value=data_points[1].value)
      nt = data_points[1].time + self.interval.ub  # when this point will expire, or if there is a point to add:

    if len(self.data_points_to_add) > 0:
      nt = min(nt, data_points_to_add[0].time + self.interval.lb)

    return nt


  def compute_qv(self, new_data_point):
    """
      This function computes the quantitative valuation when the new point arrives and returns a time point T at
      which point the quantitative valuation can change even if a new point is not received.
      The caller function is RESPONSIBLE for calling this with a new data point with time upper bounded bt T. Otherwise
      this one will produce an error.

      new_data_point: A DataPoint instance
    """
    # Arithmetic operator:
    if self.op in ['<', '>']:
      self.qv = self.param - new_data_point.value[self.metric] if self.op == '<' else new_data_point.value[self.metric] - self.param
      self.nt = U.MAX_VAL # never expire
      return  self.nt

    # Boolean operator
    if U.is_boolean(self.op):
      if self.op == '&' or self.op == '|':
        tr = self.right_node.compute_qv(new_data_point)
        tl = self.left_node.compute_qv(new_data_point)
        self.nt = min(tr,tl)
        if self.op == '&':
          self.qv = min(self.left_node.qv, self.right_node.qv)
        else:
          self.qv = max(self.left_node.qv, self.right_node.qv)
      if self.op == '!':
        self.left_node.compute_qv(new_data_point)
        self.nt = self.left_node.nt
        self.qv = -1*self.left_node.qv

      return self.nt

    # Temporal operator
    if self.op == 'P' or self.op == 'A':
      if self.nt < new_data_point.time:
        raise Exception('The valuation is changed between the last time stamp and the current one')

      # First compute the quantitative valuation for the inner formula:
      self.left_node.compute_qv(new_data_point)
      new_data_point = DataPoint(time=new_data_point.time,value=self.left_node.qv)


      if len(self.data_points_to_add) == 0 or self.data_points_to_add[-1].value != new_data_point.value:
        self.data_points_to_add.append(new_data_point)

      # Remove an expired point
      if len(self.data_points) > 1 and self.data_points[1].time <= new_data_point.time - self.interval.ub:
        z = self.data_points.pop(0)

      last_dropped_point = None
      if self.data_points_to_add[0].time <= new_data_point.time - self.interval.lb:
        decreasing = self.op == 'P'
        last_dropped_point = self._process_ordered_list(self.data_points, self.data_points_to_add.pop(0), decreasing)
        # while self._remove_condition():
        #   last_dropped_point = self.data_points.pop() # discard a data point
        # if len(self.data_points) > 0 and self.data_points[-1].value == self.data_points_to_add[0].value:
        #   self.data_points_to_add.pop(0) # just pop, do not add since it has the same value.
        # else:
        #   self.data_points.append(self.data_points_to_add.pop(0))

      self.qv = self.data_points[0].value if len(self.data_points) > 0 else U.MIN_VAL
      # Compute nt now:
      nt = self.compute_nt(self.data_points, self.data_points_to_add, last_dropped_point)
      self.nt = min(nt, self.left_node.nt) # Pick the minimum of this one and the inner formula. If the valuation
                                                # of the inner formula changes, this one can also change.
      return self.nt

    elif self.op == 'S':
      if self.nt < new_data_point.time:
        raise Exception('The valuation is changed between the last time stamp and the current one')

      self.left_node.compute_qv(new_data_point)
      self.right_node.compute_qv(new_data_point)

      xn = DataPoint(time=new_data_point.time, value=self.left_node.qv)
      yn = DataPoint(time=new_data_point.time, value=self.right_node.qv)

      # Remove expired point:
      if len(self.data_points) > 1 and self.data_points[1].time <= new_data_point.time - self.interval.ub:
        self.data_points.pop(0)
      if len(self.data_points_r) > 1 and self.data_points_r[1].time <= new_data_point.time - self.interval.ub:
        self.data_points_r.pop(0)

      # Add them to the to_add lists:
      if len(self.data_points_to_add) == 0 or self.data_points_to_add[-1].value != xn.value:
        self.data_points_to_add.append(xn)
      if len(self.data_points_to_add_r) == 0 or self.data_points_to_add_r[-1].value != yn.value:
        self.data_points_to_add_r.append(yn)


      # If the point to add is within the range:
      if self.data_points_to_add[0].time <= new_data_point.time - self.interval.lb:
        process_y = False
        process_x = False
        xn = self.data_points_to_add.pop(0)
        yn = self.data_points_to_add_r.pop(0)
        # 3 cases to check, 3 values, 6 way to order, first 2: xn, yn, qv and yn,xn,qv
        if min(xn.value, yn.value) >= self.qv or (not self.data_points or not self.data_points_r):
          self.data_points = [xn] # forget the rest
          self.data_points_r = [yn] # forget the rest

          # self.qv = min(xn.value, yn.value)
        elif self.qv >= xn.value:
          self.data_points = [xn]
          if yn.value >= xn.value:
            # 2: qv,yn,xn and yn,qv,xn
            self.data_points_r = [yn]
          else:
            # 1: qv,xn,yn
            process_y = True
          # self.qv = xn.value
        elif xn.value > self.qv >= yn.value:
          # 1: xn,qv,yn
          process_y = True
          process_x = True
          # self.qv = min(self.data_points[0].value, self.data_points_r[0].value) # only changes when a point expires

        else:
          Exception('This should not be possible.')


        dropped_x = self._process_ordered_list(self.data_points, xn, decreasing=False) if process_x else None
        dropped_y = self._process_ordered_list(self.data_points_r, yn, decreasing=True) if process_y else None
        left_nt = min(self.compute_nt(self.data_points, self.data_points_to_add, dropped_x), self.left_node.nt)
        right_nt = min(self.compute_nt(self.data_points_r, self.data_points_to_add_r, dropped_y), self.right_node.nt)
        self.nt = min(left_nt, right_nt)
        self.qv = min(self.data_points[0].value, self.data_points_r[0].value)
        return self.nt

      return self.nt



  def clean(self):
    self.data_points = []
    self.data_points_r = []
    self.data_points_to_add = []
    self.data_points_to_add_r = []
    if U.is_operand(self.op):
      return
    self.left_node.clean()
    if U.is_binary(self.op):
      self.right_node.clean()



  def to_formula(self):
    if self.op in U.MetricOperators:
      return 'x' + str(self.metric) + ' ' + self.op + ' ' + str(self.param)

    ls = self.left_node.to_formula()
    op_str = self.op
    if not U.is_boolean(self.op):
      if self.interval:
        op_str = op_str + ' ' + str(self.interval.lb) + ' ' + str(self.interval.ub)
      else:
        op_str = op_str + ' ' +  str(0) + ' p' + self.op

    if U.is_binary(self.op):
      rs = self.right_node.to_formula()
      return '( '  + ls + ' ) ' + op_str +  ' ( '  + rs + ' )'
    else:
      return op_str + ' ( ' + ls + ' )'



  # THIS FUNCTION SHOULD NOT BE IN THIS CLASS, CREATE A NEW ONE
  def compute_for_signal(self, signal, time_points):
    """
      This function will compute the quantitative valuation along the signal.
      It will iteratively call a compute function which will incrementally compute the valuation and return a time
      value t at which point the quantitative valuation might change even if a new point is not received.
    """
    result = [] # it will be a DataPoint array.
    index = 0
    dc = len(signal)
    while index <= dc:
      nt = self.compute_qv(DataPoint(time=time_points[index], value=signal[index]))
      while index + 1 <= dc and nt <= time_points[index+1]:
        nt = self.compute_qv(DataPoint(time=nt, value=signal[index]))
      index += 1

  def compute_along_signal(self, signal, time_points=None):
    """Compute the quantitative valuation along the signal. The naive and expensive approach is to call the compute at
    every point.
    """
    # - self.signal_length
    result = [0 for _ in xrange(len(signal))]
    for i in xrange(len(result)):
      if time_points:
        result[i] = self.compute(signal, i, time_points, time_points[i])
      else:
        result[i] = self.compute(signal, i, time_points)
    return result



def _to_operand_stack(op, operand_stack):
  operand2 = operand_stack.pop()
  if U.is_binary(op[0]):
    operand1 = operand_stack.pop()
    operand_stack.append(op + ' ' + operand1 + ' ' + operand2)
  else:
    operand_stack.append(op + ' ' + operand2)


def infix_to_prefix(formula):

  op_list = formula.split()

  op_stack = []
  operand_stack = []

  op_stack.append('(')
  op_list.append(')')
  output = []

  i = 0
  while i < len(op_list):
    if U.is_operand(op_list[i]):
      # x <> real
      operand_stack.append(op_list[i] + " " + op_list[i + 1] + " " + op_list[i + 2])
      i += 3
    elif op_list[i] == '(':
      op_stack.append(op_list[i])
      i += 1
    elif op_list[i] == ')':
      top_token = op_stack.pop()
      while top_token != '(':
        _to_operand_stack(top_token, operand_stack)
        top_token = op_stack.pop()
      i += 1
    else:  # it is an operator
      op = op_list[i]
      if op in U.STLOperators:
        op = op + ' ' + op_list[i + 1] + ' ' + op_list[i + 2]  # concat its bound
        i += 2
      i += 1
      while op_stack and U.PREC[op[0]] < U.PREC[op_stack[-1][0]]:
        op2 = op_stack.pop()
        _to_operand_stack(op2, operand_stack)

      op_stack.append(op)

  while op_stack:
    op = op_stack.pop()
    _to_operand_stack(op, operand_stack)

  return operand_stack[0]